(*********************************)
(*  S T A T E   P O R T R A I T  *)
(*********************************)
(* State portrait of a linear system. *)


BeginPackage["Polynomial`StatePortrait`","Graphics`PlotField`","Graphics`Arrow`","Utilities`FilterOptions`"]
(* U S A G E *)
StatePortrait::usage = "
    StatePortrait[A_?SquareMatrixQ, boundaryconditions___List, options___?OptionQ] \n
    State portrait of a linear system. \n
    Additionally, a state trajectory for an initial state vectors {X0,X1,..} is plotted. \n
";

VectorData::usage = " Generate vectors data {{x01,x02},A.{x01,x02}} for various boundary conditions. 
            \n Option: Magnification -> Real... {{x01,x02}, A.{x01,x02}/(Magnification * minvec). 
            \n Option: DataRange -> {{x1_Real,x2},{y1,y2},{dx,dy}} ... grid points of boundary conditions. ";
            
GeneralSolution::usage = " General solution of equation {x'[t] == A.x[t], x[0] == x0} ";

VectorFieldPlot::usage = " Graphics of vectors. ";

EigenVectorsPlot::usage = " Graphics of eigenvectors. ";

EigenVectorsLinePlot::usage = "Graphics of eigenvectors line. ";

TrajectoryPlot::usage = " Graphics of trajectory. ";

BoundaryConditionPlot::usage= " Graphics of trajectory for boundary conditions. ";


(* M E S S A G E *)
General::manyarg = " Many arguments in boundaryconditions: `1`.";
General::badopt = " Bad option: `1`.";

(* O P T I O N S *)
Options[VectorData] =
    {DataRange->{{-10,10},{-10,10},{1,1}},
     SetPoint->Square, Magnification->10};

Options[VectorFieldPlot] = Options[ListPlotVectorField];
SetOptions[VectorFieldPlot,DisplayFunction->Identity,ColorFunction -> Hue];

Options[EigenVectorsPlot] = Join[{ZoomVectors->1,Primitives->{RGBColor[0,0,0],Thickness[.006]}},Options[Arrow], Options[Graphics]];
SetOptions[EigenVectorsPlot, HeadLength -> 0.02 ];   (*Arrow *)

Options[EigenVectorsLinePlot] = {Zoom->15,Primitives->{Dashing[{0.03}], Thickness[0.004]}};

Options[TrajectoryPlot] = Join[{TMax->1,PointRange->{{-10,10},{-10,10},{2,2}}},Options[ParametricPlot]];
SetOptions[TrajectoryPlot, DisplayFunction->Identity,PlotStyle -> {{GrayLevel[.5], Dashing[{0.02}]}}];

Options[BoundaryConditionPlot] = Join[{TMax->2,Arrow->Automatic},Options[ParametricPlot]];
SetOptions[BoundaryConditionPlot, DisplayFunction->Identity,PlotStyle -> {{RGBColor[1,0,0],Thickness[0.005]}}];

Options[StatePortrait] =   {OptShow->{Frame->True},
                            View->{1,1,1,1,1}, (*VectorFieldPlot,EigenVectorsPlot,EigenVectorsLinePlot,TrajectoryPlot *)
                            PrintSys->False,
                            OptVectorData->Options[VectorData],
                            OptVectorFieldPlot->Options[VectorFieldPlot],
                            OptEigenVectorsPlot->Options[EigenVectorsPlot],
                            OptEigenVectorsLinePlot->Options[EigenVectorsLinePlot],
                            OptTrajectoryPlot->Options[TrajectoryPlot],
                            OptBoundaryConditionPlot->Options[BoundaryConditionPlot]
                            };

Begin["`Private`"]

(* General solution of equation {x'[t] == A.x[t], x[0] == x0} *)
GeneralSolution[A_,x01_,x02_,t_] :=
Module[{x1,x2},
    {x1[t],x2[t]}/.
    DSolve[
        Join[
            MapThread[Equal, {{x1'[t], x2'[t]}, A.{x1[t], x2[t]}}],
            MapThread[Equal, {{x1[0], x2[0]}, {x01, x02}}]
        ],
        {x1[t], x2[t]}, t
    ]//Flatten
]


(* Generate vectors - {{x1[0],x2[0]},{x1'[0],x2'[0]}} *)
VectorData[A_,opt___?OptionQ] :=
Module[{x01, x02, ders, r, fi, derc,
        data, minvec,
        range, setpoint, magni}, (* options *)

    range = DataRange /. {opt} /. Options[VectorData];
    magni = Magnification /. {opt} /. Options[VectorData];
    setpoint = SetPoint /. {opt} /. Options[VectorData];

    ders = A.{x01, x02};
    derc = A.{r*Cos[fi], r*Sin[fi]};

    Switch[setpoint,
        Square,
            data = Flatten[
                Table[{{x01, x02}, ders}, Evaluate[Flatten[{x01, range[[1]],range[[3,1]]}], Flatten[{x02, range[[2]],range[[3,2]]}]]]
            ,1];
        ,
        Circle,
            data = Flatten[
                Table[{{r*Cos[fi], r*Sin[fi]}, derc}, Evaluate[Flatten[{r, range[[1]],range[[3,1]]}], Flatten[{fi, range[[2]]*Pi/180.,range[[3,2]]*Pi/180.}]]]
            ,1];
        ,
        _,Message[General::badopt,{SetPoint}]
    ];

    If[Not[NumberQ[magni]],
        If[magni===None,Return[data],Message[General::badopt,{Magnification}]],
        (* else *)
        minvec = Min[Complement[Sqrt[Apply[Plus, (#[[2]] - #[[1]])^2]]& /@  data, {0}]];    (* find minimal vector size *)
        Return[{#[[1]], #[[2]]/(magni*minvec)}& /@ data]     ]

];


(* P L O T    F U N C T I O N *)

(* Graphics of vectors *)
VectorFieldPlot[data_,opts___?OptionQ]:=
Module[{},
    Return[
        ListPlotVectorField[data,Flatten[{{opts},{Options[VectorFieldPlot]}}]]
    ];
]

(* Graphics  of eigenvectors *)
EigenVectorsPlot[vectors_,opts___?OptionQ]:=
Module[{zoom,primitiv},
    If[
        Element[Flatten[vectors], Reals],
        {zoom} = ZoomVectors /. {opts} /. Options[EigenVectorsPlot];
        {primitiv} = Primitives /. {opts} /. Options[EigenVectorsPlot];

        Return[
            Graphics[
                {##,
                 Arrow[{0, 0}, vectors[[1]]*zoom, FilterOptions[Arrow,Flatten[{{opts},Options[EigenVectorsPlot]}]]],
                 Arrow[{0, 0}, vectors[[2]]*zoom, FilterOptions[Arrow,Flatten[{{opts},Options[EigenVectorsPlot]}]]]
                },
                FilterOptions[Graphics,Flatten[{{opts},Options[EigenVectorsPlot]}]]
            ]&@@ Flatten[{Primitives/. Options[EigenVectorsPlot],primitiv}]
        ],
        (* else *)
        Return[Graphics[{}]];
    ]

]

(* Graphics  of eigenvectors line *)
EigenVectorsLinePlot[vectors_,opts___?OptionQ]:=
Module[{primitiv,zoom},
    {primitiv} = Primitives /. {opts} /. Options[EigenVectorsLinePlot];
    {zoom} = Zoom /. {opts} /. Options[EigenVectorsLinePlot];
    If[
        Element[Flatten[vectors], Reals],
        Return[
            Graphics[{##,
                Line[{-vectors[[1]],vectors[[1]]}*zoom],
                Line[{-vectors[[2]],vectors[[2]]}*zoom]}]&@@primitiv
        ],
        (* else *)
        Return[Graphics[{}]];
    ]

]

(* Graphics  of trajectory *)
TrajectoryPlot[generalsol_,x01_,x02_,t_, opts___?OptionQ] :=
Module[{range,tmax},
        {range} = PointRange /. {opts} /. Options[TrajectoryPlot];
        {tmax} = TMax /. {opts} /. Options[TrajectoryPlot];
        ParametricPlot[
            Evaluate[
                Flatten[
                    Table[generalsol,
                        Evaluate[Flatten[{x01, range[[1]],range[[3,1]]}], Flatten[{x02, range[[2]],range[[3,2]]}]]
                    ], 1
                ]
            ], {t, 0, tmax},
            FilterOptions[ParametricPlot,Flatten[{opts, Options[TrajectoryPlot]}]]//Evaluate
        ]//Return
];

(* Graphics of trajectory for boundary conditions *)
BoundaryConditionPlot[generalsol_,bc_,x01_,x02_,t_, opts___?OptionQ] :=
Module[{tmax,arrow,nbc,paramfce,p=10,arrs,graphicsarrow},
    {tmax} = TMax /. {opts} /. Options[BoundaryConditionPlot];
    {arrow} = Arrow /. {opts} /. Options[BoundaryConditionPlot];
    If[Dimensions[bc]==={2},nbc={bc},nbc=bc]; (* if bc = {x01,x02} only *)    
    paramfce = Apply[Function[{x01, x02}, generalsol// Evaluate], nbc, {1}]//Simplify;

    Switch[arrow,
        Automatic,
            arrs=MapThread[Arrow[##, HeadLength->0.04, HeadWidth -> 0.5, HeadCenter -> .5]&,
                 {
                    Function[Evaluate[t, paramfce]] /@ (Range[0, tmax, tmax/p]) //Flatten[#, 1]&,
                    Function[Evaluate[t, paramfce]] /@ (Range[0, tmax, tmax/p] + .01) //Flatten[#, 1]&
                 }];
            graphicsarrow = 
                Graphics[{
                    Sequence @@ (PlotStyle/.{FilterOptions[ParametricPlot,Flatten[{opts, Options[BoundaryConditionPlot]}]]})[[1]]
                    ,arrs
                }]
        ,
        None,
            graphicsarrow = Graphics[{}]
        ,
        _,Message[General::badopt,{Arrow}]
    ];

    Return[{  
        graphicsarrow,
        ParametricPlot[
            Evaluate[paramfce], {t, 0, tmax},
            FilterOptions[ParametricPlot,Flatten[{opts, Options[BoundaryConditionPlot]}]]//Evaluate
        ]
    }]
];

SquareMatrixQ[matrix_?MatrixQ] := SameQ @@ Dimensions[matrix];

(******************************)
(*  STATE PORTRAIT            *)
(******************************)
StatePortrait[A_?SquareMatrixQ,bc___List,opts___?OptionQ]:=
Module[{optvectorfieldplot,optvectordata,optvectorsplot,optvectorslineplot,opttrajectoryplot,optshow,optbcplot, (* variable for options *)
        view, printsys, (* options for StatePortrait*)
        plotlabel,plotrange,datarange,pointrange,
        data,sys,sol,x01,x02,t,
        pvf,pev,pevl,ptr,pbc},  (* variable for graphics *)

    If[ Length[{bc}]>1, Message[General::manyarg,{bc}];Return[]];

    (* read options *)
    view = View /. {opts} /. Options[StatePortrait];
    printsys = PrintSys /. {opts} /. Options[StatePortrait];

    optvectordata = OptVectorData /. {opts} /. Options[StatePortrait];
    optvectorfieldplot = OptVectorFieldPlot /. {opts} /. Options[StatePortrait];
    optvectorsplot = OptEigenVectorsPlot /. {opts} /. Options[StatePortrait];
    optvectorslineplot = OptEigenVectorsLinePlot /. {opts} /. Options[StatePortrait];
    opttrajectoryplot = OptTrajectoryPlot /. {opts} /. Options[StatePortrait];
    optshow = OptShow /. {opts} /. Options[StatePortrait];
    optbcplot = OptBoundaryConditionPlot /. {opts} /. Options[StatePortrait];

    data = VectorData[A,##]& @@ optvectordata;
    sys = Expand[Eigensystem[A]];
    sol = GeneralSolution[A,x01,x02,t];

    If[printsys,
        Print["Eigen values: ",Expand[sys[[1]]]];
        Print["Eigen vectors: ",Expand[sys[[2]]]];
    ];

    plotlabel =
        StringJoin[
            StringInsert["Eigen values:    \n", ToString[StandardForm[sys[[1]]]], 16],
            StringInsert["Eigen vectors:    ", ToString[StandardForm[sys[[2]]]], 16]
        ];

    datarange = (DataRange /. Flatten[{optvectordata,OptVectorData /.Options[StatePortrait]}]//Evaluate);        
    plotrange = datarange[[{1, 2}]] + datarange[[{1, 2}]]/5;

    pointrange = datarange; 
    pointrange[[3]] = 2*pointrange[[3]]; (* 2x ridsi sit *)
    opttrajectoryplot = Flatten[{opttrajectoryplot,{PointRange->pointrange}}];
    
    
    (* generate graphics *)        
    If[view[[1]]>0, pvf = VectorFieldPlot[data,optvectorfieldplot], pvf = Graphics[{}]];
    If[view[[2]]>0, pev = EigenVectorsPlot[sys[[2]],optvectorsplot], pev = Graphics[{}]];
    If[view[[3]]>0, pevl = EigenVectorsLinePlot[sys[[2]],optvectorslineplot], pevl = Graphics[{}]];
    If[view[[4]]>0, ptr = TrajectoryPlot[sol, x01, x02, t,opttrajectoryplot], ptr = Graphics[{}]];
    If[view[[5]]>0 && Length[{bc}]===1, pbc = BoundaryConditionPlot[sol, bc, x01, x02, t,optbcplot], pbc = Graphics[{}]];

    (* show all graphics *)
    Show[{pvf,pevl,ptr,pbc,pev},DisplayFunction->$DisplayFunction,
        ##
    ]& @@ Flatten[{optshow,{PlotLabel->plotlabel,PlotRange->plotrange},OptShow/.Options[StatePortrait]}]
]; 
End[ ]

EndPackage[] 